// Energy Router - Diverting excess solar PV energy to HWS storage.
// Also incorporating measurement and display via LCD screen, and
// reporting of data via MySensors.

// Cary Wintle - Feb 2017

// Revision D - Request existing values from controller on startup,
//              store rollover values in EEPROM.
// Revision C - Remove PhaseCal, add Vrms calculation.
// Revision B - Integer maths, interrupt-driven sampling.
// Revision A - Float maths, random sampling.

// Based on:
// [ Mk2i PV Router rev 6a by Robin Emley (calypso_rae on Open Energy Monitor Forum) 
// (https://openenergymonitor.org/emon/mk2) ]
// and
// [ MySensors (http://www.mysensors.org) created by Henrik Ekblad <henrik.ekblad@mysensors.org>
// * Copyright (C) 2013-2015 Sensnology AB
// * Full contributor list: https://github.com/mysensors/Arduino/graphs/contributors ]

// Modified to use an external phase angle controller which has 0-5V control signal.
// (Control signal is generated using filtered PWM.)
// MySensors library modified to work with MQTT OpenHAB
// Modified to display data locally on 20x4 LCD display

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// version 2 as published by the Free Software Foundation.

// TODO: Refine algorithm for EnergyBucket -> PowerLevel calculation
// TODO: Fix glitches when display updates

// Pin assignments
// ---------------
#define outputPinForModule 9
const byte voltageSensor = 5;          // A2 is for the voltage sensor (ADC5)
const byte currentSensor_solar = 7;    // A0 is CT2 for solar current (ADC7)
const byte currentSensor_grid = 6;     // A1 is CT1 for grid current (ADC6)

// MySensors configuration
// -----------------------
#define SN "Energy Router"  // Software name
#define SV "0.4"            // Version number
//#define MY_DEBUG
#define MY_NODE_ID 2
#define MY_RADIO_RFM69
#define MY_RFM69_FREQUENCY RF69_433MHZ
#define MY_IS_RFM69HW
#define MY_RF69_IRQ_PIN 7
#define MY_RF69_IRQ_NUM 4
#define MY_RF69_SPI_CS 10
#define MY_DEFAULT_TX_LED_PIN 8
#include <MySensors.h>

// Software configuration
// ----------------------
#define EEPROM_RO_DATA EEPROM_LOCAL_CONFIG_ADDRESS + 1 // Where the midnight rollover values start
#define ADC_TIMER_PERIOD 125      // uS (determines the sampling rate / amount of idle time)
#define CYCLES_PER_SECOND 50      // Local mains frequency
#define SWEETZONE_IN_JOULES 3600  // 0.001 kWh = 3600 Joules
#define ANTI_CREEP_LIMIT 5        // in Joules per mains cycle (has no effect when set to 0)
#define RF_SEND_PERIOD 10         // seconds between MySensors reports
#define TIME_REQ_TO 60000L        // Milliseconds to wait for response from HA controller
const byte startUpPeriod = 3;     // in seconds, to allow LP filter to settle
const int DCoffset_I = 512;       // nominal mid-point value of ADC @ x1 scale

// Calibration values
//-------------------
const float powerCal_grid = 0.12317;  // for CT1
const float powerCal_solar = 0.12317;  // for CT2
#define V_CAL 0.8298 // for Voltage sensor

// LCD Display
//------------
#include <Wire.h>
#include <LiquidCrystal_PCF8574.h>
LiquidCrystal_PCF8574 lcd(0x38); // set the LCD address

// Physical constants, please do not change!
#define JOULES_PER_WATT_HOUR 3600 //  (1 Wh = 0.001 kWh = 3600 Joules)
enum polarities {NEGATIVE, POSITIVE};

#include <TimerOne.h> // Used to time the ADC samples
#include <EEPROM.h> // Used to store rollover values

// MySensors Messages 
// ------------------
//(NB: *HP* denotes high priority messages that are sent every reporting interval
// To save processing time the remaining messages are sent in a sequence, one at a time per reporting interval.

#define GridChildID 1 // Grid import sensor child
MyMessage gridwattMsg(GridChildID, V_WATT); // *HP*
MyMessage gridkwhMsg(GridChildID, V_KWH);

#define GridExportChildID 2 // Grid export sensor child
MyMessage exportkwhMsg(GridExportChildID, V_KWH);
MyMessage gridVoltsMsg(GridExportChildID, V_VOLTAGE);

#define SolarChildID 3 // Id of the solar power sensor child
MyMessage solarwattMsg(SolarChildID, V_WATT); // *HP*
MyMessage solarkwhMsg(SolarChildID, V_KWH);

#define BPChildID 4 // ID of the bypass dimmer control child
MyMessage switchMsg(BPChildID, V_STATUS);
MyMessage dimmerMsg(BPChildID, V_PERCENTAGE);

// Global variables
// ----------------
byte cycleCount = 0;                    // counts mains cycles
long triggerThreshold_long;        // for determining when the trigger may be safely armed
long energyInBucket_long;          // in Integer Energy Units
long capacityOfEnergyBucket_long;  // depends on powerCal, frequency & the 'sweetzone' size.
long upperEnergyBucket_long;
long lowerEnergyBucket_long;
long DCoffset_V_long;              // <--- for LPF
long DCoffset_V_min;               // <--- for LPF
long DCoffset_V_max;               // <--- for LPF
int realPower_grid_int, realPower_solar_int; // Current power readings +ve is generation
long solarEnergyRecent_IEU = 0; // Hi-res accumulator of limited range
long gridEnergyRecent_IEU = 0; // Hi-res accumulator of limited range
unsigned int importEnergyTotal_Wh = 0; // WattHour register of 63K range
unsigned int exportEnergyTotal_Wh = 0; // WattHour register of 63K range
unsigned int solarEnergyTotal_Wh = 0; // WattHour register of 63K range
long IEU_per_Wh_grid,IEU_per_Wh_solar; // depends on powerCal & frequency.
float IEU_to_Joules_conversion; // for displaying
long sumP_atGrid; // for datalogging
long sumP_atSolar; // for datalogging
long Vrms;
unsigned int samplesDuringDatalogPeriod; // for datalogging
long antiCreepLimit_inIEUperMainsCycle;

// for interaction between the main processor and the ISRs 
volatile boolean dataReady = false;
volatile int sampleI_grid,sampleI_solar,sampleV;

byte MessageSequence = 0; // Sequence state counter for messaging cycle

unsigned long Time; // Time in seconds since 1970
unsigned long Next10sTime; // Time stored by 10s update routine
unsigned long MidnightTime = 2500; // Time at next midnight (initial value delays lockup if no time reply)
volatile boolean timeValid = false; // A correct time has been received from the controller
volatile boolean doRollover = false; // Flag that the midnight rollover needs to happen
unsigned long timeRequestTimeout; // millis() when we need to request time again
struct  { 
  uint8_t Second; 
  uint8_t Minute; 
  uint8_t Hour; 
  uint8_t Wday;   // day of week, sunday is day 1
  uint8_t Day;
  uint8_t Month; 
  uint8_t Year;   // offset from 1970; 
}   tm; // Time structure
byte secondCount = 0;

boolean beyondStartUpPhase = false;

boolean Importing = false;

boolean BypassSwitch = false; // Bypasses the power level calculation when requested by MySensors
byte RequestLevel = 0; // 0-100% power request from the MySensors network

// For power module control
unsigned long PowerLevel; // 10-bit value to send to PWM output to control power output of module

// Custom characters for LCD display
#define WHChar 0
byte WHCharArray[8] = { // Watt-hour character
  0b10001,0b10101,0b11011,0b10001,0b01010,0b01110,0b01010,0b01010 };

#define TodaySolarChar 1
byte TodaySolarCharArray[8] = { // Today Solar Character
  0b01110,0b00100,0b00100,0b10001,0b01110,0b01010,0b01110,0b10001 };

#define YestSolarChar 2
byte YestSolarCharArray[8] = { // Yesterday Solar Character
  0b01010,0b00100,0b00100,0b10001,0b01110,0b01010,0b01110,0b10001 };

#define ExpChar 3
byte ExpCharArray[8] = { // Export Character
  0b00011,0b00111,0b01111,0b11111,0b01111,0b00111,0b00011,0b00000 };

#define ImpChar 4
byte ImpCharArray[8] = { // Import Character
  0b11000,0b11100,0b11110,0b11111,0b11110,0b11100,0b11000,0b00000 };

// In-built characters for LCD display
byte TodayImpChar = 0x3E;
byte TodayExpChar = 0x3C;
byte YestImpChar = 0x7E;
byte YestExpChar = 0x7F;
byte SunChar = 0x2A;


// -----------------------------------------------------------
void setup() {  

  // Energy maths setup
  capacityOfEnergyBucket_long = (long)SWEETZONE_IN_JOULES * CYCLES_PER_SECOND * (1/powerCal_grid);
  energyInBucket_long = capacityOfEnergyBucket_long * 0.45; // for rapid start up
  upperEnergyBucket_long = 2600L * CYCLES_PER_SECOND * (1/powerCal_grid);
  lowerEnergyBucket_long = 1000L * CYCLES_PER_SECOND * (1/powerCal_grid);
  // for display of energy bucket's level
  IEU_to_Joules_conversion = (float)SWEETZONE_IN_JOULES / capacityOfEnergyBucket_long;
  antiCreepLimit_inIEUperMainsCycle = (float)ANTI_CREEP_LIMIT * (1/powerCal_solar);

  IEU_per_Wh_grid = (long)JOULES_PER_WATT_HOUR * CYCLES_PER_SECOND * (1/powerCal_grid);
  IEU_per_Wh_solar = (long)JOULES_PER_WATT_HOUR * CYCLES_PER_SECOND * (1/powerCal_solar);

  DCoffset_V_long = 512L * 256; // nominal mid-point value of ADC @ x256 scale  
  DCoffset_V_min = (long)(512L - 100) * 256; // mid-point of ADC minus a working margin
  DCoffset_V_max = (long)(512L + 100) * 256; // mid-point of ADC plus a working margin

  // LCD Display setup and initial layout
  lcd.begin(20, 4); // initialize the lcd 20 characters 4 lines
  lcd.clear();
  lcd.setBacklight(255);
  // Load in the custom characters
  lcd.createChar(WHChar, WHCharArray);
  lcd.createChar(TodaySolarChar, TodaySolarCharArray);
  lcd.createChar(YestSolarChar, YestSolarCharArray);
  lcd.createChar(ExpChar, ExpCharArray);
  lcd.createChar(ImpChar, ImpCharArray);
  // Setup the fixed display elements
  lcd.setCursor(6,0);
  lcd.write("W");
  lcd.setCursor(8,0);
  lcd.write(SunChar);
  lcd.setCursor(14,0);
  lcd.write("W");
  lcd.setCursor(19,0);
  lcd.write("V");
  lcd.setCursor(0,1);
  lcd.write((uint8_t)TodaySolarChar);
  lcd.setCursor(6,1);
  lcd.write((uint8_t)WHChar);
  lcd.setCursor(8,1);
  lcd.write((uint8_t)YestSolarChar);
  lcd.setCursor(14,1);
  lcd.write((uint8_t)WHChar);
  lcd.setCursor(16,1);
  lcd.write("N");
  lcd.setCursor(19,1);
  lcd.write("%");
  lcd.setCursor(0,2);
  lcd.write((uint8_t)TodayImpChar);
  lcd.setCursor(6,2);
  lcd.write((uint8_t)WHChar);
  lcd.setCursor(8,2);
  lcd.write((uint8_t)YestImpChar);
  lcd.setCursor(14,2);
  lcd.write((uint8_t)WHChar);
  lcd.setCursor(0,3);
  lcd.write((uint8_t)TodayExpChar);
  lcd.setCursor(6,3);
  lcd.write((uint8_t)WHChar);
  lcd.setCursor(8,3);
  lcd.write((uint8_t)YestExpChar);
  lcd.setCursor(14,3);
  lcd.write((uint8_t)WHChar);

  // Read previous day's values from EEPROM
  byte low = EEPROM.read(EEPROM_RO_DATA);
  byte high = EEPROM.read(EEPROM_RO_DATA+1);
  solarEnergyTotal_Wh = word(high,low);
  low = EEPROM.read(EEPROM_RO_DATA+2);
  high = EEPROM.read(EEPROM_RO_DATA+3);
  importEnergyTotal_Wh = word(high,low);
  low = EEPROM.read(EEPROM_RO_DATA+4);
  high = EEPROM.read(EEPROM_RO_DATA+5);
  exportEnergyTotal_Wh = word(high,low);
  // Output to display
  char LCDmsg[6];
  sprintf(LCDmsg, "%5.1u", solarEnergyTotal_Wh);
  lcd.setCursor(9,1);
  lcd.print(LCDmsg);
  sprintf(LCDmsg, "%5.1u", importEnergyTotal_Wh);
  lcd.setCursor(9,2);
  lcd.print(LCDmsg);
  sprintf(LCDmsg, "%5.1u", exportEnergyTotal_Wh);
  lcd.setCursor(9,3);
  lcd.print(LCDmsg);

  // Request time and current data from the controller
  // NB: MySensors core library modified to send "1" on requests (MQTT)
  request(BPChildID, V_STATUS, 0); // Request the current bypass switch state
  wait(TIME_REQ_TO, 1, V_STATUS); // Wait for the response
  request(BPChildID, V_PERCENTAGE, 0); // Request the current power level
  wait(TIME_REQ_TO, 1, V_PERCENTAGE); // Wait for the response
  request(GridChildID, V_KWH, 0); // Request the current state of accumulators
  wait(TIME_REQ_TO, 1, V_KWH); // Wait for the response
  request(GridExportChildID, V_KWH, 0);
  wait(TIME_REQ_TO, 1, V_KWH); // Wait for the response
  request(SolarChildID, V_KWH, 0);
  wait(TIME_REQ_TO, 1, V_KWH); // Wait for the response
  requestTime();

  // Set up the ADC to be triggered by a hardware timer of fixed duration  
  ADCSRA  = (1<<ADPS0)+(1<<ADPS1)+(1<<ADPS2);  // Set the ADC's clock to system clock / 128
  ADCSRA |= (1 << ADEN);                 // Enable ADC

  Timer1.initialize(ADC_TIMER_PERIOD);   // set Timer1 interval
  Timer1.attachInterrupt(timerIsr);    // declare timerIsr() as interrupt service routine
}

// ------------------------------------------------------------
void presentation() {
    // Send the sketch version information to the gateway and Controller
    sendSketchInfo(SN, SV);

    // Present the child sensors
    present(GridChildID, S_POWER);
    present(SolarChildID, S_POWER);
    present(GridExportChildID, S_POWER);
    present(BPChildID, S_DIMMER);
}

// ------------------------------------------------------------
void timerIsr(void)
{                                         
  static unsigned char sample_index = 0;

  switch(sample_index) {
    case 0:
      sampleV = ADC;                    // store the ADC value (this one is for Voltage)
      ADMUX = 0x40 + currentSensor_solar;  // set up the next conversion, which is for Solar Current
      ADCSRA |= (1<<ADSC);              // start the ADC
      sample_index++;                   // increment the control flag
      dataReady = true;                 // ADC values can now be processed
      break;
    case 1:
      sampleI_solar = ADC;               // store the ADC value (this one is for Solar Current)
      ADMUX = 0x40 + currentSensor_grid;  // set up the next conversion, which is for Grid Current
      ADCSRA |= (1<<ADSC);              // start the ADC
      sample_index++;                   // increment the control flag
      break;
    case 2:
      sampleI_grid = ADC;               // store the ADC value (this one is for Grid Current)
      ADMUX = 0x40 + voltageSensor;         // set up the next conversion, which is for Voltage
      ADCSRA |= (1<<ADSC);              // start the ADC
      sample_index =0;                   // reset the control flag
      break;
    default:
      sample_index = 0;                 // to prevent lockup (should never get here)      
  }
}

// ------------------------------------------------------------
void loop() {

    if (dataReady) {   // flag is set after every set of ADC conversions
      dataReady = false; // reset the flag
      allGeneralProcessing(); // executed once for each set of V&I samples
    }
    
    // Midnight rollover
    if (doRollover) { 
      if (timeValid && tm.Hour == 0) { // Check that we're past midnight
        char LCDmsg[6];
        sprintf(LCDmsg, "%5.1u", solarEnergyTotal_Wh);
        lcd.setCursor(9,1);
        lcd.print(LCDmsg);
        sprintf(LCDmsg, "%5.1u", importEnergyTotal_Wh);
        lcd.setCursor(9,2);
        lcd.print(LCDmsg);
        sprintf(LCDmsg, "%5.1u", exportEnergyTotal_Wh);
        lcd.setCursor(9,3);
        lcd.print(LCDmsg);
        // Save the values to EEPROM
        EEPROM.write(EEPROM_RO_DATA,lowByte(solarEnergyTotal_Wh));
        EEPROM.write(EEPROM_RO_DATA+1,highByte(solarEnergyTotal_Wh));
        EEPROM.write(EEPROM_RO_DATA+2,lowByte(importEnergyTotal_Wh));
        EEPROM.write(EEPROM_RO_DATA+3,highByte(importEnergyTotal_Wh));
        EEPROM.write(EEPROM_RO_DATA+4,lowByte(exportEnergyTotal_Wh));
        EEPROM.write(EEPROM_RO_DATA+5,highByte(exportEnergyTotal_Wh));
        // Reset the accumulators
        solarEnergyTotal_Wh = 0;
        importEnergyTotal_Wh = 0;
        exportEnergyTotal_Wh = 0;
        solarEnergyRecent_IEU = 0;
        doRollover = false;
      }
    }
    
    //Update some of the display once per second
    if (cycleCount >= (int)(CYCLES_PER_SECOND-1)) {
      cycleCount = 0;
      secondCount++;

      // Update the time
      Time++;
      if (timeValid) {
        // Check for midnight
        if (Time > (MidnightTime + 1)) { // It's around midnight, but let's update the clock to be sure
          doRollover = true; // Flag that rollover has to happen (prevents it happening more than once)
          timeValid = false;
          timeRequestTimeout = millis() + TIME_REQ_TO;
          requestTime();
        }
        tm.Second++;
        if (tm.Second > 59) {
          tm.Second = 0;
          tm.Minute++;
          if (tm.Minute > 59) {
            tm.Minute = 0;
            tm.Hour++;
            if (tm.Hour < 23) {
              timeValid = false;
              timeRequestTimeout = millis() + TIME_REQ_TO;
              requestTime();   // Once an hour, update the clock from the controller
            }                  // (Not after 10pm though, or midnight rollover never happens)
          }
        }
      }
      else {
        if (millis() > timeRequestTimeout) { // We got no response from our timeRequest within timeout period
          timeValid = false;
          timeRequestTimeout = millis() + TIME_REQ_TO;
          requestTime(); // Re-send request
        }
      }
      
      // calculate the average power at the supply point
      long realPower_grid_long = sumP_atGrid / samplesDuringDatalogPeriod; 
      realPower_grid_int = realPower_grid_long * powerCal_grid; // where import is +ve
      if (realPower_grid_int > 0)
        Importing = true;
      else
        Importing = false;
      // calculate the average power from the solar array
      long realPower_solar_long = sumP_atSolar / samplesDuringDatalogPeriod;
      if (realPower_solar_long < 0)
        realPower_solar_long = 0; // Solar generation is never negative
      realPower_solar_int = realPower_solar_long * powerCal_solar; // generation is +ve
      // Reset the datalogging interval
      samplesDuringDatalogPeriod = 0;
      sumP_atGrid = 0;
      sumP_atSolar = 0;
      
      lcd.setCursor(0,0);
      if (Importing)
        lcd.write((uint8_t)ImpChar);
      else
        lcd.write((uint8_t)ExpChar);
      char LCDmsg[6];
      sprintf(LCDmsg, "%5.1d", realPower_grid_int);
      lcd.setCursor(1,0);
      lcd.print(LCDmsg);
      sprintf(LCDmsg, "%5.1d", realPower_solar_int);
      lcd.setCursor(9,0);
      lcd.print(LCDmsg);
      sprintf(LCDmsg, "%03u", Vrms);
      lcd.setCursor(16,0);
      lcd.print(LCDmsg);
      lcd.setCursor(17,1);
      int Level = map(PowerLevel,0,1023,0,99);
      sprintf(LCDmsg, "%02u", Level);
      lcd.print(LCDmsg[0]);
      lcd.print(LCDmsg[1]);
      float BucketDisplay = energyInBucket_long * IEU_to_Joules_conversion;
      lcd.setCursor(16,2);
      dtostrf(BucketDisplay, 4, 0, LCDmsg);
      lcd.print(LCDmsg);
    } // End of once-per-second routine

    // Update kwh displays once per ten seconds
    if (Time > Next10sTime) {
      char LCDmsg[6];
      Next10sTime = Time + 9;
      sprintf(LCDmsg, "%5.1u", solarEnergyTotal_Wh);
      lcd.setCursor(1,1);
      lcd.print(LCDmsg);
      sprintf(LCDmsg, "%5.1u", importEnergyTotal_Wh);
      lcd.setCursor(1,2);
      lcd.print(LCDmsg);
      sprintf(LCDmsg, "%5.1u", exportEnergyTotal_Wh);
      lcd.setCursor(1,3);
      lcd.print(LCDmsg);
      // Update time display
      lcd.setCursor(16,3);
      if (tm.Hour < 10)
        lcd.print("0");
      lcd.print(tm.Hour);
      if (tm.Minute < 10)
        lcd.print("0");
      lcd.print(tm.Minute);
    }

    if (secondCount > (RF_SEND_PERIOD - 1)) { // Time to send some MyS messages
      secondCount = 0;
      // MySensors message cycle
      switch (MessageSequence) {
        case 0: // kwh Imported
          send(gridkwhMsg.set((int)importEnergyTotal_Wh));
          send(gridwattMsg.set((int)realPower_grid_int));
          send(solarwattMsg.set((int)realPower_solar_int));
          MessageSequence++;
          break;
        case 1: // kwh Exported
          send(exportkwhMsg.set((int)exportEnergyTotal_Wh));
          send(gridwattMsg.set((int)realPower_grid_int));
          send(solarwattMsg.set((int)realPower_solar_int));
          MessageSequence++;
          break;
        case 2: // kwh Solar
          send(solarkwhMsg.set((int)solarEnergyTotal_Wh));
          send(gridwattMsg.set((int)realPower_grid_int));
          send(solarwattMsg.set((int)realPower_solar_int));
          MessageSequence++;
          break;
        case 3: // grid Volts
          send(gridVoltsMsg.set((int)Vrms));
          send(gridwattMsg.set((int)realPower_grid_int));
          send(solarwattMsg.set((int)realPower_solar_int));
          MessageSequence++;
          break;
        case 4: // Bypass Switch
          send(switchMsg.set(BypassSwitch));
          send(gridwattMsg.set((int)realPower_grid_int));
          send(solarwattMsg.set((int)realPower_solar_int));
          MessageSequence++;
          break;
        case 5: // Bypass Power Level
          send(dimmerMsg.set(map(PowerLevel,0,1023,0,100)));
          send(gridwattMsg.set((int)realPower_grid_int));
          send(solarwattMsg.set((int)realPower_solar_int));
          MessageSequence = 0;
          break;
        default:
          MessageSequence = 0;
        break;
      }
    }
    
} // end of loop()

// ------------------------------------------------------------
void allGeneralProcessing()
{
  static int samplesDuringThisCycle; // for normalising the power in each mains cycle
  static long sumP_grid; // for per-cycle summation of 'real power' 
  static long sumP_solar; // for per-cycle summation of 'real power' 
  static enum polarities polarityOfLastSampleV; // for zero-crossing detection
  static long cumVdeltasThisCycle_long; // for the LPF which determines DC offset (voltage)
  static long lastSampleVminusDC_long; //    for the phaseCal algorithm
  static long sumV; // accumulator for Vrms
 
  // remove DC offset from the raw voltage sample by subtracting the accurate value 
  // as determined by a LP filter.
  long sampleVminusDC_long = ((long)sampleV<<8) - DCoffset_V_long;
  int sampleVminusDC = (int)sampleV - 512;

  //Vrms accumulator
  sumV += (long)sampleVminusDC * (long)sampleVminusDC;

  // determine the polarity of the latest voltage sample
  enum polarities polarityNow;   
  if(sampleVminusDC_long > 0) { 
    polarityNow = POSITIVE; }
  else { 
    polarityNow = NEGATIVE; }

  if (polarityNow == POSITIVE) 
  { 
    if (beyondStartUpPhase)
    {     
      if (polarityOfLastSampleV != POSITIVE)
      {
        // This is the start of a new +ve half cycle (just after the zero-crossing point)
        cycleCount++;  

        // Calculate the Vrms
        Vrms = V_CAL * sqrt(sumV / samplesDuringThisCycle);
        sumV = 0;
    
        // Calculate the real power and energy during the last whole mains cycle.
        //
        // sumP contains the sum of many individual calculations of instantaneous power.  In  
        // order to obtain the average power during the relevant period, sumP must first be 
        // divided by the number of samples that have contributed to its value.
        //
        // The next stage would normally be to apply a calibration factor so that real power 
        // can be expressed in Watts.  That's fine for floating point maths, but it's not such
        // a good idea when integer maths is being used.  To keep the numbers large, and also 
        // to save time, calibration of power is omitted at this stage.  Real Power (stored as 
        // a 'long') is therefore (1/powerCal) times larger than the actual power in Watts.
        //
        
        long realPower_grid = sumP_grid / samplesDuringThisCycle; // proportional to Watts
        long realPower_solar = sumP_solar / samplesDuringThisCycle; // proportional to Watts
   
        // Next, the energy content of this power rating needs to be determined.  Energy is 
        // power multiplied by time, so the next step is normally to multiply the measured
        // value of power by the time over which it was measured.
        //   Instanstaneous power is calculated once every mains cycle. When integer maths is 
        // being used, a repetitive power-to-energy conversion seems an unnecessary workload.  
        // As all sampling periods are of similar duration, it is more efficient simply to 
        // add all of the power samples together, and note that their sum is actually 
        // CYCLES_PER_SECOND greater than it would otherwise be.
        //   Although the numerical value itself does not change, I thought that a new name 
        // may be helpful so as to minimise confusion.  
        //   The 'energy' variable below is CYCLES_PER_SECOND * (1/powerCal) times larger than 
        // the actual energy in Joules.
        // positive values of grid are importing, negative are exporting
        
        long realEnergy_grid = realPower_grid; 
        long realEnergy_solar = realPower_solar;

         if (realEnergy_solar < antiCreepLimit_inIEUperMainsCycle) {
            realEnergy_solar = 0; } // to avoid 'creep'
        
   
        // Energy contributions from the grid connection point (CT1) are summed in an 
        // accumulator which is known as the energy bucket.  The purpose of the energy bucket 
        // is to mimic the operation of the supply meter.  The range over which energy can 
        // pass to and fro without loss or charge to the user is known as its 'sweet-zone'.
        // The capacity of the energy bucket is set to this same value within setup().
        //    
        // The latest contribution can now be added to this energy bucket
        energyInBucket_long -= realEnergy_grid;   // Subtract to make positive power (import) reduce the bucket
         
        // Apply max and min limits to bucket's level.  This is to ensure correct operation
        // when conditions change, i.e. when import changes to export, and vici versa.
        //
        if (energyInBucket_long > capacityOfEnergyBucket_long) { 
          energyInBucket_long = capacityOfEnergyBucket_long; } 
        else         
        if (energyInBucket_long < 0) {
          energyInBucket_long = 0; }  

          // ***** Start of segment to control power module - FIXME: REDO THIS!!
          if (!BypassSwitch) { //If the bypass switch is off, set PowerLevel
            // Zero power if energy level is below lower threshold
            if (energyInBucket_long <= lowerEnergyBucket_long)
              PowerLevel = 0;
            else  
            // Full power if energy level is above upper threshold
            if (energyInBucket_long >= upperEnergyBucket_long) 
              PowerLevel = 1023;
            else {
              // determine the appropriate power level for the bucket's level
              // (Quick and dirty scaling using 2^10 division)
              PowerLevel = energyInBucket_long - lowerEnergyBucket_long;
              PowerLevel = PowerLevel >> 10;
              if (PowerLevel > 1023)
                PowerLevel = 1023;
            }
          } // If the bypass switch is on, PowerLevel will be set when the MySensors message arrives
          analogWrite(outputPinForModule, PowerLevel);
          // ***** End of segment to control power module
  
        // The latest contribution needs to be added to an accumulator which operates with maximum precision.
        solarEnergyRecent_IEU += realEnergy_solar;
        gridEnergyRecent_IEU += realEnergy_grid;
    
        // Whole Wh are then recorded separately
        if (solarEnergyRecent_IEU > IEU_per_Wh_solar)
        {
          solarEnergyRecent_IEU -= IEU_per_Wh_solar;
          solarEnergyTotal_Wh++;
        } 
        
        if (gridEnergyRecent_IEU > 0) {
          if (gridEnergyRecent_IEU > IEU_per_Wh_grid)
          {
            gridEnergyRecent_IEU -= IEU_per_Wh_grid;
            importEnergyTotal_Wh++;
          } 
        }
        else {
          if (-gridEnergyRecent_IEU > IEU_per_Wh_grid)
          {
            gridEnergyRecent_IEU += IEU_per_Wh_grid;
            exportEnergyTotal_Wh++;
          } 
          
        }

        // clear the per-cycle accumulators for use in this new mains cycle.  
        samplesDuringThisCycle = 0;
        sumP_grid = 0;
        sumP_solar = 0;

      } // end of processing that is specific to the first Vsample in each +ve half cycle 
  
    }
    else
    {  
      // wait until the DC-blocking filters have had time to settle
      if(millis() > startUpPeriod * 1000) 
      {
        beyondStartUpPhase = true;
        sumP_grid = 0;
        sumP_solar = 0;
        samplesDuringThisCycle = 0;
      }
    }
    
  } // end of processing that is specific to samples where the voltage is positive
  
  else // the polatity of this sample is negative
  {     
    if (polarityOfLastSampleV != NEGATIVE)
    {
      // This is the start of a new -ve half cycle (just after the zero-crossing point)      
      // which is a convenient point to update the Low Pass Filter for DC-offset removal
      //
      long previousOffset = DCoffset_V_long;
      DCoffset_V_long = previousOffset + (cumVdeltasThisCycle_long>>6); // faster than * 0.01
      cumVdeltasThisCycle_long = 0;
      
      // To ensure that the LPF will always start up correctly when 240V AC is available, its
      // output value needs to be prevented from drifting beyond the likely range of the 
      // voltage signal.  This avoids the need to use a HPF as was done for initial Mk2 builds.
      //
      if (DCoffset_V_long < DCoffset_V_min) {
        DCoffset_V_long = DCoffset_V_min; }
      else  
      if (DCoffset_V_long > DCoffset_V_max) {
        DCoffset_V_long = DCoffset_V_max; }
           
    } // end of processing that is specific to the first Vsample in each -ve half cycle
  } // end of processing that is specific to samples where the voltage is positive
  
  // processing for EVERY pair of samples
  //
  // First, deal with the power at the grid connection point (as measured via CT1)
  // remove most of the DC offset from the current sample (the precise value does not matter)
  long sampleIminusDC_grid = ((long)(sampleI_grid - DCoffset_I))<<8;
                      
  // calculate the grid "real power" in this sample pair and add to the accumulated sum
  long filtV_div4 = sampleVminusDC_long>>2;  // reduce to 16-bits (now x64, or 2^6)
  long filtI_div4 = sampleIminusDC_grid>>2; // reduce to 16-bits (now x64, or 2^6)
  long instP = filtV_div4 * filtI_div4;  // 32-bits (now x4096, or 2^12)
  instP = instP>>12;     // scaling is now x1, as for Mk2 (V_ADC x I_ADC)       
  sumP_grid +=instP; // cumulative power, scaling as for Mk2 (V_ADC x I_ADC)
  sumP_atGrid +=instP; // cumulative power, for datalogging
  
  // Now deal with the solar power (as measured via CT2)
  // remove most of the DC offset from the current sample (the precise value does not matter)
  long sampleIminusDC_solar = ((long)(sampleI_solar - DCoffset_I))<<8;
                                                               
  // calculate the solar "real power" in this sample pair and add to the accumulated sum
  filtI_div4 = sampleIminusDC_solar>>2; // reduce to 16-bits (now x64, or 2^6)
  instP = filtV_div4 * filtI_div4;  // 32-bits (now x4096, or 2^12)
  instP = instP>>12;     // scaling is now x1, as for Mk2 (V_ADC x I_ADC)       
  sumP_solar +=instP; // cumulative power, scaling as for Mk2 (V_ADC x I_ADC)
  sumP_atSolar +=instP; // cumulative power, for datalogging
  
  samplesDuringThisCycle++;
  samplesDuringDatalogPeriod++;
  
  // store items for use during next loop
  cumVdeltasThisCycle_long += sampleVminusDC_long; // for use with LP filter
  polarityOfLastSampleV = polarityNow;  // for identification of half cycle boundaries
}

// ------------------------------------------------------------
void receive(const MyMessage &message) { // Receive messages from the gateway
  if (message.type == V_STATUS) {
    BypassSwitch = atoi(message.data);
    lcd.setCursor(16,1);
    if (BypassSwitch) {
      PowerLevel = map(RequestLevel, 0,100,0,1023); // Map it to the full range of the output module
      lcd.print("B"); }
    else
        lcd.print("N"); 
  }
  if (message.type == V_PERCENTAGE) {
    RequestLevel = atoi(message.data); // Sent as 0-100 percent
    if (BypassSwitch) 
      PowerLevel = map(RequestLevel, 0,100,0,1023); // Map it to the full range of the output module
  }
  if (message.type == V_KWH) { // These messages update the accumulators after a restart
    switch (message.sensor) {
      case GridChildID:
        importEnergyTotal_Wh = atoi(message.data);
        break;
      case GridExportChildID:
        exportEnergyTotal_Wh = atoi(message.data);
        break;
      case SolarChildID:
        solarEnergyTotal_Wh = atoi(message.data);
        break;
    }
  }
}

// ------------------------------------------------------------
void receiveTime(unsigned long ts) {
  Time = ts; // Time is received as seconds since 1970, adjusted for local time zone
  breakTime(); // Updaste the time structure
  // Calculate when the next midnight will be (in Time seconds)
  MidnightTime = ((long)(23 - tm.Hour) * 3600) + ((long)(59 - tm.Minute) * 60) + (long)(60 - tm.Second) + Time;
  // Update the display
  lcd.setCursor(16,3);
  if (tm.Hour < 10)
    lcd.print("0");
  lcd.print(tm.Hour);
  if (tm.Minute < 10)
    lcd.print("0");
  lcd.print(tm.Minute);
  timeValid = true;
}

// ------------------------------------------------------------
void breakTime() {
// Break the current Time (in seconds since 1970) into time components (Y,M,D,H,M,S)
// Note that year is offset from 1970
// Result is stored in the tm structure

// leap year calulator expects year argument as years offset from 1970
#define LEAP_YEAR(Y)     ( ((1970+Y)>0) && !((1970+Y)%4) && ( ((1970+Y)%100) || !((1970+Y)%400) ) )

static  const uint8_t monthDays[]={31,28,31,30,31,30,31,31,30,31,30,31}; 
// API starts months from 1, this array starts from 0

  uint32_t tempTime = Time;
  tm.Second = tempTime % 60;
  tempTime /= 60; // now it is minutes
  tm.Minute = tempTime % 60;
  tempTime /= 60; // now it is hours
  tm.Hour = tempTime % 24;
  /* (uncomment the rest of this if we ever care about days, months and years)
  tempTime /= 24; // now it is days
  tm.Wday = ((tempTime + 4) % 7) + 1;  // Sunday is day 1 
  
  uint8_t year = 0;  
  unsigned long days = 0;
  while((unsigned)(days += (LEAP_YEAR(year) ? 366 : 365)) <= tempTime) {
    year++;
  }
  tm.Year = year; // year is offset from 1970 
  
  days -= LEAP_YEAR(year) ? 366 : 365;
  tempTime  -= days; // now it is days in this year, starting at 0
  
  days=0;
  uint8_t month=0;
  uint8_t monthLength=0;
  for (month=0; month<12; month++) {
    if (month==1) { // february
      if (LEAP_YEAR(year)) {
        monthLength=29;
      } else {
        monthLength=28;
      }
    } else {
      monthLength = monthDays[month];
    }
    
    if (tempTime >= monthLength) {
      tempTime -= monthLength;
    } else {
        break;
    }
  }
  tm.Month = month + 1;  // jan is month 1  
  tm.Day = tempTime + 1;     // day of month */
}
